# [[file:../../org/pyretechnics.org::surface-fire-imports][surface-fire-imports]]
import cython
import cython as cy
if cython.compiled:
    from cython.cimports.libc.math import sqrt, pow, log, exp
    from cython.cimports.pyretechnics.cy_types import \
        vec_xy, vec_xyz, fcatarr, fclaarr, FuelModel, ProjectedVectors, FireBehaviorMin, \
        FireBehaviorMax, SpreadBehavior
    import cython.cimports.pyretechnics.conversion as conv
    import cython.cimports.pyretechnics.vector_utils as vu
else:
    from math import sqrt, pow, log, exp
    from pyretechnics.py_types import \
        vec_xy, vec_xyz, fcatarr, fclaarr, FuelModel, ProjectedVectors, FireBehaviorMin, \
        FireBehaviorMax, SpreadBehavior
    import pyretechnics.conversion as conv
    import pyretechnics.vector_utils as vu
# surface-fire-imports ends here
# [[file:../../org/pyretechnics.org::surface-fire-utility-functions][surface-fire-utility-functions]]
@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def __dotp_in_category(x_ij: fclaarr, y_ij: fclaarr) -> fcatarr:
    """
    Util: dot-product between fuel size class vectors, within each fuel category (dead/live).
    """
    return (
        (
            (x_ij[0] * y_ij[0]) +
            (x_ij[1] * y_ij[1]) +
            (x_ij[2] * y_ij[2]) +
            (x_ij[3] * y_ij[3])
        ),
        (
            (x_ij[4] * y_ij[4]) +
            (x_ij[5] * y_ij[5])
        )
    )


@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def __dotp_categories(x_i: fcatarr, y_i: fcatarr) -> cy.float:
    return (
        (x_i[0] * y_i[0]) +
        (x_i[1] * y_i[1])
    )
# surface-fire-utility-functions ends here
# [[file:../../org/pyretechnics.org::surface-fire-common-intermediate-calculations][surface-fire-common-intermediate-calculations]]
@cy.cfunc
@cy.exceptval(check=False)
def calc_surface_area_to_volume_ratio(f_i: fcatarr, f_ij: fclaarr, sigma: fclaarr) -> cy.float:
    sigma_prime: fcatarr = __dotp_in_category(f_ij, sigma)
    return __dotp_categories(f_i, sigma_prime)


@cy.cfunc
@cy.exceptval(check=False)
def calc_packing_ratio(w_o: fclaarr, rho_p: fclaarr, delta: cy.float) -> cy.float:
    if (delta > 0.0):
        rho_p_inv: fclaarr = ( # TODO: OPTIM pre-compute
            1.0/rho_p[0],
            1.0/rho_p[1],
            1.0/rho_p[2],
            1.0/rho_p[3],
            1.0/rho_p[4],
            1.0/rho_p[5]
        )
        beta: fcatarr = __dotp_in_category(w_o, rho_p_inv)
        return (beta[0] + beta[1]) / delta
    else:
        return 0.0


@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def calc_optimum_packing_ratio(sigma_prime: cy.float) -> cy.float:
    return (3.348 * pow(sigma_prime, -0.8189)) if (sigma_prime > 0.0) else 1.0
# surface-fire-common-intermediate-calculations ends here
# [[file:../../org/pyretechnics.org::surface-fire-reaction-intensity][surface-fire-reaction-intensity]]
@cy.cfunc
@cy.exceptval(check=False)
def calc_mineral_damping_coefficients(f_ij: fclaarr, S_e: fclaarr) -> fcatarr:
    S_e_i: fcatarr = __dotp_in_category(f_ij, S_e)
    (S_e_0, S_e_1) = S_e_i
    return (
        0.174 * pow(S_e_0, -0.19) if (S_e_0 > 0.0) else 1.0,
        0.174 * pow(S_e_1, -0.19) if (S_e_1 > 0.0) else 1.0
    )


@cy.cfunc
@cy.exceptval(check=False)
def __cat_moisture_damping_coefficient(M_f: cy.float, M_x: cy.float) -> cy.float:
    if (M_x > 0.0):
        r_M : cy.float = min(1.0, M_f / M_x)
        r_M2: cy.float = r_M * r_M
        r_M3: cy.float = r_M2 * r_M
        return 1.0 - (2.59 * r_M) + (5.11 * r_M2) - (3.52 * r_M3)
    else:
        return 0.0


@cy.cfunc
@cy.exceptval(check=False)
def calc_moisture_damping_coefficients(f_ij: fclaarr, M_f: fclaarr, M_x: fclaarr) -> fcatarr:
    M_f_i: fcatarr = __dotp_in_category(f_ij, M_f)
    M_x_i: fcatarr = __dotp_in_category(f_ij, M_x)
    return (
        __cat_moisture_damping_coefficient(M_f_i[0], M_x_i[0]),
        __cat_moisture_damping_coefficient(M_f_i[1], M_x_i[1])
    )


@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def calc_low_heat_content(f_ij: fclaarr, h: fclaarr) -> fcatarr:
    return __dotp_in_category(f_ij, h)


@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def calc_net_fuel_loading(g_ij: fclaarr, w_o: fclaarr, S_T: fclaarr) -> fcatarr:
    return (
        (
            g_ij[0] * w_o[0] * (1.0 - S_T[0]) +
            g_ij[1] * w_o[1] * (1.0 - S_T[1]) +
            g_ij[2] * w_o[2] * (1.0 - S_T[2]) +
            g_ij[3] * w_o[3] * (1.0 - S_T[3])
        ),
        (
            g_ij[4] * w_o[4] * (1.0 - S_T[4]) +
            g_ij[5] * w_o[5] * (1.0 - S_T[5])
        )
    )


@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def calc_heat_per_unit_area(eta_S_i: fcatarr, eta_M_i: fcatarr, h_i: fcatarr, W_n_i: fcatarr) -> cy.float:
    return (
        (W_n_i[0] * h_i[0] * eta_M_i[0] * eta_S_i[0]) +
        (W_n_i[1] * h_i[1] * eta_M_i[1] * eta_S_i[1])
    )


@cy.cfunc
@cy.exceptval(check=False)
def calc_optimum_reaction_velocity(sigma_prime: cy.float, beta: cy.float, beta_op: cy.float) -> cy.float:
    # Albini 1976 replaces 1 / (4.774 * (sigma_prime ** 0.1) - 7.27)
    A: cy.float = (133.0 * pow(sigma_prime, -0.7913)) if (sigma_prime > 0.0) else 0.0
    B: cy.float = pow(sigma_prime, 1.5)
    C: cy.float = beta / beta_op
    # Maximum reaction velocity (1/min)
    Gamma_prime_max: cy.float = B / (495.0 + 0.0594 * B)
    # Optimum reaction velocity (1/min)
    return Gamma_prime_max * pow(C, A) * exp(A * (1.0 - C))


@cy.cfunc
@cy.exceptval(check=False)
def calc_reaction_intensity(moisturized_fuel_model: FuelModel,
                            sigma_prime           : cy.float,
                            beta                  : cy.float,
                            beta_op               : cy.float) -> cy.float:
    # Unpack the FuelModel fields
    w_o : fclaarr = moisturized_fuel_model.w_o
    h   : fclaarr = moisturized_fuel_model.h
    S_T : fclaarr = moisturized_fuel_model.S_T
    S_e : fclaarr = moisturized_fuel_model.S_e
    M_x : fclaarr = moisturized_fuel_model.M_x
    M_f : fclaarr = moisturized_fuel_model.M_f
    f_ij: fclaarr = moisturized_fuel_model.f_ij
    g_ij: fclaarr = moisturized_fuel_model.g_ij
    # Derive intermediate quantities
    eta_S_i    : fcatarr  = calc_mineral_damping_coefficients(f_ij, S_e)
    eta_M_i    : fcatarr  = calc_moisture_damping_coefficients(f_ij, M_f, M_x)
    h_i        : fcatarr  = calc_low_heat_content(f_ij, h)                             # (Btu/lb)
    W_n_i      : fcatarr  = calc_net_fuel_loading(g_ij, w_o, S_T)                      # (lb/ft^2)
    Btus       : cy.float = calc_heat_per_unit_area(eta_S_i, eta_M_i, h_i, W_n_i)      # (Btu/ft^2)
    Gamma_prime: cy.float = calc_optimum_reaction_velocity(sigma_prime, beta, beta_op) # (1/min)
    # Calculate reaction intensity
    return Btus * Gamma_prime # (Btu/ft^2/min)
# surface-fire-reaction-intensity ends here
# [[file:../../org/pyretechnics.org::surface-fire-propagating-flux-ratio][surface-fire-propagating-flux-ratio]]
@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def calc_propagating_flux_ratio(sigma_prime: cy.float, beta: cy.float) -> cy.float:
    return exp((0.792 + 0.681 * sqrt(sigma_prime)) * (beta + 0.1)) / (192.0 + 0.2595 * sigma_prime)
# surface-fire-propagating-flux-ratio ends here
# [[file:../../org/pyretechnics.org::surface-fire-heat-source-no-wind-no-slope][surface-fire-heat-source-no-wind-no-slope]]
@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def calc_heat_source(I_R: cy.float, xi: cy.float) -> cy.float:
    return I_R * xi
# surface-fire-heat-source-no-wind-no-slope ends here
# [[file:../../org/pyretechnics.org::surface-fire-oven-dry-fuel-bed-bulk-density][surface-fire-oven-dry-fuel-bed-bulk-density]]
@cy.cfunc
@cy.exceptval(check=False)
def calc_ovendry_bulk_density(w_o: fclaarr, delta: cy.float) -> cy.float:
    if (delta > 0.0):
        w_o_sum: cy.float = w_o[0] + w_o[1] + w_o[2] + w_o[3] + w_o[4] + w_o[5]
        return w_o_sum / delta
    else:
        return 0.0
# surface-fire-oven-dry-fuel-bed-bulk-density ends here
# [[file:../../org/pyretechnics.org::surface-fire-effective-heating-number-distribution][surface-fire-effective-heating-number-distribution]]
@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def __sizeclass_heating_number(sigma_i: cy.float) -> cy.float:
    return exp(-138.0 / sigma_i) if (sigma_i > 0.0) else 0.0


@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def calc_effective_heating_number_distribution(sigma: fclaarr) -> fclaarr: # OPTIM pre-compute, exp is expensive
    return (
        __sizeclass_heating_number(sigma[0]),
        __sizeclass_heating_number(sigma[1]),
        __sizeclass_heating_number(sigma[2]),
        __sizeclass_heating_number(sigma[3]),
        __sizeclass_heating_number(sigma[4]),
        __sizeclass_heating_number(sigma[5])
    )
# surface-fire-effective-heating-number-distribution ends here
# [[file:../../org/pyretechnics.org::surface-fire-heat-of-preignition-distribution][surface-fire-heat-of-preignition-distribution]]
@cy.cfunc
@cy.exceptval(check=False)
def __sizeclass_heat_of_preignition_distribution(M_f_i: cy.float) -> cy.float:
    pos1: cy.float = cy.cast(cy.float, (M_f_i > 0.0))
    return (250.0 + 1116.0 * M_f_i) * pos1 # Returns 0 unless M_f_i > 0


@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def calc_heat_of_preignition_distribution(M_f: fclaarr) -> fclaarr:
    return (
        __sizeclass_heat_of_preignition_distribution(M_f[0]),
        __sizeclass_heat_of_preignition_distribution(M_f[1]),
        __sizeclass_heat_of_preignition_distribution(M_f[2]),
        __sizeclass_heat_of_preignition_distribution(M_f[3]),
        __sizeclass_heat_of_preignition_distribution(M_f[4]),
        __sizeclass_heat_of_preignition_distribution(M_f[5])
    )
# surface-fire-heat-of-preignition-distribution ends here
# [[file:../../org/pyretechnics.org::surface-fire-heat-sink][surface-fire-heat-sink]]
@cy.cfunc
@cy.exceptval(check=False)
def calc_heat_sink(f_i: fcatarr, f_ij: fclaarr, rho_b: cy.float, epsilon_ij: fclaarr, Q_ig_ij: fclaarr) -> cy.float:
    effective_heat_of_preignition_i: fcatarr = (
        (
            (f_ij[0] * epsilon_ij[0] * Q_ig_ij[0]) +
            (f_ij[1] * epsilon_ij[1] * Q_ig_ij[1]) +
            (f_ij[2] * epsilon_ij[2] * Q_ig_ij[2]) +
            (f_ij[3] * epsilon_ij[3] * Q_ig_ij[3])
        ),
        (
            (f_ij[4] * epsilon_ij[4] * Q_ig_ij[4]) +
            (f_ij[5] * epsilon_ij[5] * Q_ig_ij[5])
        )
    )
    effective_heat_of_preignition: cy.float = __dotp_categories(f_i, effective_heat_of_preignition_i)
    return rho_b * effective_heat_of_preignition
# surface-fire-heat-sink ends here
# [[file:../../org/pyretechnics.org::surface-fire-spread-rate-no-wind-no-slope][surface-fire-spread-rate-no-wind-no-slope]]
@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def calc_spread_rate(heat_source: cy.float, heat_sink: cy.float) -> cy.float:
    return heat_source / heat_sink if (heat_sink > 0.0) else 0.0
# surface-fire-spread-rate-no-wind-no-slope ends here
# [[file:../../org/pyretechnics.org::surface-fire-intensity-functions][surface-fire-intensity-functions]]
@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def calc_residence_time(sigma_prime: cy.float) -> cy.float:
    """
    Returns the residence time (total burning time) of fuel (min) given:
    - sigma_prime :: ft^2/ft^3 (surface area to volume ratio)
    """
    return 384.0 / sigma_prime if (sigma_prime > 0.0) else 0.0


@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def calc_flame_depth(spread_rate: cy.float, residence_time: cy.float) -> cy.float:
    """
    Returns the depth, or front-to-back distance, of the actively flaming zone
    of a free-spreading fire (ft) given:
    - spread_rate    :: ft/min (orthogonal to the fireline)
    - residence_time :: min
    """
    return spread_rate * residence_time


@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def calc_fireline_intensity(reaction_intensity: cy.float, flame_depth: cy.float) -> cy.float:
    """
    Returns the rate of heat release per unit of fire edge (Btu/ft/s) given:
    - reaction_intensity :: Btu/ft^2/min
    - flame_depth        :: ft
    """
    return (reaction_intensity * flame_depth) / 60.0


@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def calc_flame_length(fireline_intensity: cy.float) -> cy.float:
    """
    Returns the average flame length (m) given:
    - fireline_intensity :: kW/m
    """
    return 0.07747042253266703 * pow(fireline_intensity, 0.46)


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def calc_areal_heat_output(spread_rate: cy.float, fireline_intensity: cy.float) -> cy.float:
    """
    Returns the heat per unit area (kJ/m^2) given:
    - spread_rate        :: m/min
    - fireline_intensity :: kW/m
    """
    return 60.0 * fireline_intensity / spread_rate if spread_rate > 0.0 else 0.0
# surface-fire-intensity-functions ends here
# [[file:../../org/pyretechnics.org::surface-fire-max-effective-wind-speed][surface-fire-max-effective-wind-speed]]
@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def calc_max_effective_wind_speed(reaction_intensity: cy.float) -> cy.float:
    return 0.9 * reaction_intensity
# surface-fire-max-effective-wind-speed ends here
# [[file:../../org/pyretechnics.org::surface-fire-slope-factor-function][surface-fire-slope-factor-function]]
@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def get_phi_S(surface_fire_min: FireBehaviorMin, slope: cy.float) -> cy.float:
    return surface_fire_min._phiS_G * (slope * slope)
# surface-fire-slope-factor-function ends here
# [[file:../../org/pyretechnics.org::surface-fire-wind-factor-function][surface-fire-wind-factor-function]]
@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def get_phi_W(surface_fire_min: FireBehaviorMin, midflame_wind_speed: cy.float) -> cy.float:
    return surface_fire_min._phiW_scalr * pow(midflame_wind_speed, surface_fire_min._phiW_expnt)
# surface-fire-wind-factor-function ends here
# [[file:../../org/pyretechnics.org::surface-fire-wind-speed-function][surface-fire-wind-speed-function]]
@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def get_wind_speed(surface_fire_min: FireBehaviorMin, phi_W: cy.float) -> cy.float:
    return surface_fire_min._ws_scalr * pow(phi_W, surface_fire_min._ws_expnt)
# surface-fire-wind-speed-function ends here
# [[file:../../org/pyretechnics.org::surface-fire-behavior-no-wind-no-slope][surface-fire-behavior-no-wind-no-slope]]
@cy.cfunc
@cy.exceptval(check=False)
def make_surface_fire_min(base_spread_rate        : cy.float,
                          base_fireline_intensity : cy.float,
                          max_effective_wind_speed: cy.float,
                          B                       : cy.float,
                          C                       : cy.float,
                          F                       : cy.float,
                          beta                    : cy.float) -> FireBehaviorMin:
    _phiS_G: cy.float = 0.0
    if (beta > 0.0):
        _phiS_G = 5.275 * pow(beta, -0.3)

    _phiW_scalr: cy.float = 0.0
    _phiW_expnt: cy.float = 0.0
    if (F > 0.0):
        _phiW_scalr = (C / F) * pow(conv.m_to_ft(1.0), B)
        _phiW_expnt = B

    _ws_scalr: cy.float = 0.0
    _ws_expnt: cy.float = 0.0
    if (B > 0.0):
        B_inverse: cy.float = 1.0 / B
        _ws_scalr = conv.ft_to_m(pow((F / C), B_inverse))
        _ws_expnt = B_inverse

    return FireBehaviorMin(
        base_spread_rate,
        base_fireline_intensity,
        max_effective_wind_speed,
        _phiS_G,
        _phiW_scalr,
        _phiW_expnt,
        _ws_scalr,
        _ws_expnt)


@cy.ccall
@cy.exceptval(check=False)
def calc_surface_fire_behavior_no_wind_no_slope(moisturized_fuel_model: FuelModel,
                                                spread_rate_adjustment: cy.float = 1.0) -> FireBehaviorMin:
    """
    Given these inputs:
    - moisturized_fuel_model :: FuelModel struct of fuel model and fuel moisture properties
      - delta :: ft                                        (fuel depth)
      - w_o   :: lb/ft^2                                   (ovendry fuel loading)
      - rho_p :: lb/ft^3                                   (ovendry particle density)
      - sigma :: ft^2/ft^3                                 (fuel particle surface-area-to-volume ratio)
      - h     :: Btu/lb                                    (fuel particle low heat content)
      - S_T   :: lb minerals/lb ovendry weight             (fuel particle total mineral content)
      - S_e   :: lb silica-free minerals/lb ovendry weight (fuel particle effective mineral content)
      - M_x   :: lb moisture/lb ovendry weight             (fuel particle moisture of extinction)
      - M_f   :: lb moisture/lb ovendry weight             (fuel particle moisture content)
      - f_ij  :: %                                         (percent load per size class)
      - f_i   :: %                                         (percent load per category)
      - g_ij  :: %                                         (percent load per size class - Albini_1976_FIREMOD, page 20)
    - spread_rate_adjustment :: unitless float (1.0 for no adjustment)

    return a FireBehaviorMin struct containing these keys:
    - base_spread_rate         :: m/min
    - base_fireline_intensity  :: kW/m
    - max_effective_wind_speed :: m/min
    - _phiS_G                  :: intermediate value for computing phi_S (unitless)
    - _phiW_scalr              :: intermediate value for computing phi_W (unitless)
    - _phiW_expnt              :: intermediate value for computing phi_W (unitless)
    - _ws_scalr                :: intermediate value for computing effective_wind_speed (m/min)
    - _ws_expnt                :: intermediate value for computing effective_wind_speed (unitless)
    """
    # Unpack fuel model values
    delta: cy.float = moisturized_fuel_model.delta
    w_o  : fclaarr  = moisturized_fuel_model.w_o
    rho_p: fclaarr  = moisturized_fuel_model.rho_p
    sigma: fclaarr  = moisturized_fuel_model.sigma
    M_f  : fclaarr  = moisturized_fuel_model.M_f
    f_ij : fclaarr  = moisturized_fuel_model.f_ij
    f_i  : fcatarr  = moisturized_fuel_model.f_i
    # Calculate base spread rate (no wind, no slope)
    sigma_prime: cy.float = calc_surface_area_to_volume_ratio(f_i, f_ij, sigma)
    beta       : cy.float = calc_packing_ratio(w_o, rho_p, delta)
    beta_op    : cy.float = calc_optimum_packing_ratio(sigma_prime)
    I_R        : cy.float = calc_reaction_intensity(moisturized_fuel_model, sigma_prime, beta, beta_op) # Btu/ft^2/min
    xi         : cy.float = calc_propagating_flux_ratio(sigma_prime, beta)
    heat_source: cy.float = calc_heat_source(I_R, xi)                             # Btu/ft^2/min
    rho_b      : cy.float = calc_ovendry_bulk_density(w_o, delta)                 # lb/ft^3
    epsilon_ij : fclaarr  = calc_effective_heating_number_distribution(sigma)
    Q_ig_ij    : fclaarr  = calc_heat_of_preignition_distribution(M_f)            # Btu/lb
    heat_sink  : cy.float = calc_heat_sink(f_i, f_ij, rho_b, epsilon_ij, Q_ig_ij) # Btu/ft^3
    R0         : cy.float = calc_spread_rate(heat_source, heat_sink)              # ft/min
    # Calculate base fireline intensity (no wind, no slope)
    t_res: cy.float = calc_residence_time(sigma_prime)  # min
    D_A  : cy.float = calc_flame_depth(R0, t_res)       # ft
    I_s  : cy.float = calc_fireline_intensity(I_R, D_A) # Btu/ft/s
    # Pre-compute values related to wind and slope
    U_eff_max: cy.float = calc_max_effective_wind_speed(I_R) # ft/min
    B        : cy.float = 0.02526 * pow(sigma_prime, 0.54)
    C        : cy.float = 7.47 * exp(-0.133 * pow(sigma_prime, 0.55))
    E        : cy.float = 0.715 * exp(-3.59 * (sigma_prime * 1e-4))
    F        : cy.float = pow((beta / beta_op), E)
    # Return no-wind-no-slope surface fire behavior values
    base_spread_rate        : cy.float = conv.ft_to_m(R0 * spread_rate_adjustment)
    base_fireline_intensity : cy.float = conv.Btu_ft_s_to_kW_m(I_s * spread_rate_adjustment)
    max_effective_wind_speed: cy.float = conv.ft_to_m(U_eff_max)
    return make_surface_fire_min(
        base_spread_rate,
        base_fireline_intensity,
        max_effective_wind_speed,
        B,
        C,
        F,
        beta
    )
# surface-fire-behavior-no-wind-no-slope ends here
# [[file:../../org/pyretechnics.org::midflame-wind-speed][midflame-wind-speed]]
@cy.cfunc
@cy.exceptval(check=False)
def calc_wind_adjustment_factor(fuel_bed_depth: cy.float, canopy_height: cy.float, canopy_cover: cy.float) -> cy.float:
    """
    Return the wind adjustment factor (unitless) given these inputs:
    - fuel_bed_depth :: ft
    - canopy_height  :: ft
    - canopy_cover   :: 0-1
    """
    if (canopy_cover > 0.0) and (canopy_height > 0.0):
        # sheltered: equation 2 based on CC and CH, CR=1 (Andrews 2012)
        A: cy.float = sqrt((canopy_cover / 3.0) * canopy_height)
        B: cy.float = log((20.0 + 0.36 * canopy_height) / (0.13 * canopy_height))
        return 0.555 / (A * B)
    elif (fuel_bed_depth > 0.0):
        # unsheltered: equation 6 H_F = H (Andrews 2012)
        A: cy.float = log((20.0 + 0.36 * fuel_bed_depth) / (0.13 * fuel_bed_depth))
        return 1.83 / A # 1.83 truncated from 1.8328795184533409
    else:
        # non-burnable fuel model
        return 0.0


@cy.ccall
@cy.exceptval(check=False)
def calc_midflame_wind_speed(wind_speed_20ft: cy.float,
                             fuel_bed_depth : cy.float,
                             canopy_height  : cy.float,
                             canopy_cover   : cy.float) -> cy.float:
    """
    Return the midflame wind speed (S) given these inputs:
    - wind_speed_20ft :: S
    - fuel_bed_depth  :: ft
    - canopy_height   :: ft
    - canopy_cover    :: 0-1
    """
    wind_adj_factor: cy.float = calc_wind_adjustment_factor(fuel_bed_depth, canopy_height, canopy_cover)
    return wind_speed_20ft * wind_adj_factor
# midflame-wind-speed ends here
# [[file:../../org/pyretechnics.org::surface-fire-combine-wind-and-slope-vectors][surface-fire-combine-wind-and-slope-vectors]]
@cy.cfunc
@cy.exceptval(check=False)
def project_wind_and_slope_vectors_3d(wind_speed        : cy.float,
                                      downwind_direction: cy.float,
                                      slope             : cy.float,
                                      upslope_direction : cy.float) -> ProjectedVectors:
    """
    Given these inputs:
    - wind_speed         :: S
    - downwind_direction :: degrees clockwise from North
    - slope              :: rise/run
    - upslope_direction  :: degrees clockwise from North

    return a ProjectedVectors struct containing these keys:
    - wind_vector_3d  :: (x: S, y: S, z: S)
    - slope_vector_3d :: (x, y, z)
    """
    # Convert wind and slope vectors from azimuthal to cartesian coordinates
    wind_vector_2d : vec_xy = conv.azimuthal_to_cartesian(wind_speed, downwind_direction)
    slope_vector_2d: vec_xy = conv.azimuthal_to_cartesian(slope, upslope_direction)
    # Project wind and slope vectors onto the slope-tangential plane
    wind_vector_3d : vec_xyz = vu.to_slope_plane(wind_vector_2d, slope_vector_2d)
    slope_vector_3d: vec_xyz = vu.to_slope_plane(slope_vector_2d, slope_vector_2d)
    return ProjectedVectors(wind_vector_3d, slope_vector_3d)


@cy.cfunc
@cy.exceptval(check=False)
def get_phi_E(wind_vector_3d: vec_xyz, slope_vector_3d: vec_xyz, phi_W: cy.float, phi_S: cy.float) -> vec_xyz:
    # Convert wind and slope vectors to unit vectors on the slope-tangential plane
    w_S: vec_xyz = vu.as_unit_vector_3d(wind_vector_3d)  if phi_W > 0.0 else wind_vector_3d
    u_S: vec_xyz = vu.as_unit_vector_3d(slope_vector_3d) if phi_S > 0.0 else slope_vector_3d
    # Create the 3D slope-tangential phi_W, phi_S, and phi_E vectors
    phi_W_3d: vec_xyz = vu.scale_3d(phi_W, w_S)
    phi_S_3d: vec_xyz = vu.scale_3d(phi_S, u_S)
    phi_E_3d: vec_xyz = vu.add_3d(phi_W_3d, phi_S_3d)
    return phi_E_3d
# surface-fire-combine-wind-and-slope-vectors ends here
# [[file:../../org/pyretechnics.org::surface-fire-eccentricity][surface-fire-eccentricity]]
# TODO: Change model from str to enumerated type
@cy.cfunc
def surface_length_to_width_ratio(effective_wind_speed: cy.float, model: str = "behave") -> cy.float:
    """
    Calculate the length_to_width_ratio of the surface fire front given:
    - effective_wind_speed :: m/min (aligned with the slope-tangential plane)
    - model                :: "rothermel" or "behave" (Optional)
    """
    effective_wind_speed_mph = conv.m_min_to_mph(effective_wind_speed)
    # Select formula by model
    if model == "rothermel":
        return 1.0 + 0.25 * effective_wind_speed_mph

    elif model == "behave":
        return min(8.0,
                   0.936 * exp(0.1147 * effective_wind_speed_mph)
                   +
                   0.461 * exp(-0.0692 * effective_wind_speed_mph)
                   -
                   0.397)

    else:
        raise ValueError("Invalid input: model must be 'rothermel' or 'behave'.")


@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def surface_fire_eccentricity(length_to_width_ratio: cy.float) -> cy.float:
    """
    Calculate the eccentricity (E) of the surface fire front using eq. 8 from
    Albini and Chase 1980 given:
    - L/W :: (1: circular spread, > 1: elliptical spread)
    """
    return sqrt((length_to_width_ratio * length_to_width_ratio) - 1.0) / length_to_width_ratio
# surface-fire-eccentricity ends here
# [[file:../../org/pyretechnics.org::surface-fire-behavior-max][surface-fire-behavior-max]]
@cy.cfunc
@cy.exceptval(check=False)
def maybe_limit_wind_speed(use_wind_limit  : cy.bint,
                           max_wind_speed  : cy.float,
                           surface_fire_min: FireBehaviorMin,
                           phi_E_magnitude : cy.float) -> tuple[cy.float, cy.float]:
    """
    Given these inputs:
    - use_wind_limit   :: boolean
    - max_wind_speed   :: m/min
    - surface_fire_min :: FireBehaviorMin struct of no-wind-no-slope surface fire behavior values
      - base_spread_rate         :: m/min
      - base_fireline_intensity  :: kW/m
      - max_effective_wind_speed :: m/min
      - _phiS_G                  :: intermediate value for computing phi_S (unitless)
      - _phiW_scalr              :: intermediate value for computing phi_W (unitless)
      - _phiW_expnt              :: intermediate value for computing phi_W (unitless)
      - _ws_scalr                :: intermediate value for computing effective_wind_speed (m/min)
      - _ws_expnt                :: intermediate value for computing effective_wind_speed (unitless)
    - phi_E_magnitude  :: unitless

    return a tuple with these fields:
    - limited_wind_speed :: m/min
    - limited_phi_E      :: unitless
    """
    effective_wind_speed: cy.float = get_wind_speed(surface_fire_min, phi_E_magnitude)
    if (use_wind_limit and effective_wind_speed > max_wind_speed):
        return (
            max_wind_speed,
            get_phi_W(surface_fire_min, max_wind_speed),
        )
    else:
        return (
            effective_wind_speed,
            phi_E_magnitude,
        )


@cy.ccall
@cy.exceptval(check=False)
def calc_surface_fire_behavior_max(surface_fire_min      : FireBehaviorMin,
                                   midflame_wind_speed   : cy.float,
                                   upwind_direction      : cy.float,
                                   slope                 : cy.float,
                                   aspect                : cy.float,
                                   use_wind_limit        : cy.bint = True,
                                   surface_lw_ratio_model: str = "behave") -> FireBehaviorMax:
    """
    Given these inputs:
    - surface_fire_min       :: FireBehaviorMin struct of no-wind-no-slope surface fire behavior values
      - base_spread_rate         :: m/min
      - base_fireline_intensity  :: kW/m
      - max_effective_wind_speed :: m/min
      - _phiS_G                  :: intermediate value for computing phi_S (unitless)
      - _phiW_scalr              :: intermediate value for computing phi_W (unitless)
      - _phiW_expnt              :: intermediate value for computing phi_W (unitless)
      - _ws_scalr                :: intermediate value for computing effective_wind_speed (m/min)
      - _ws_expnt                :: intermediate value for computing effective_wind_speed (unitless)
    - midflame_wind_speed    :: m/min
    - upwind_direction       :: degrees clockwise from North
    - slope                  :: rise/run
    - aspect                 :: degrees clockwise from North
    - use_wind_limit         :: boolean (Optional)
    - surface_lw_ratio_model :: "rothermel" or "behave" (Optional)

    return a FireBehaviorMax struct containing these keys:
    - max_fire_type          :: 1 (surface)
    - max_spread_rate        :: m/min
    - max_spread_direction   :: (x, y, z) unit vector
    - max_fireline_intensity :: kW/m
    - max_flame_length       :: m
    - length_to_width_ratio  :: unitless (1: circular spread, > 1: elliptical spread)
    - eccentricity           :: unitless (0: circular spread, > 0: elliptical spread)
    - critical_spread_rate   :: m/min
    """
    # Unpack no-wind-no-slope surface fire behavior values
    spread_rate       : cy.float = surface_fire_min.base_spread_rate
    fireline_intensity: cy.float = surface_fire_min.base_fireline_intensity
    max_wind_speed    : cy.float = surface_fire_min.max_effective_wind_speed
    # Reverse the provided wind and slope directions
    downwind_direction: cy.float = conv.opposite_direction(upwind_direction)
    upslope_direction : cy.float = conv.opposite_direction(aspect)
    # Project wind and slope vectors onto the slope-tangential plane
    vectors        : ProjectedVectors = project_wind_and_slope_vectors_3d(midflame_wind_speed,
                                                                          downwind_direction,
                                                                          slope,
                                                                          upslope_direction)
    wind_vector_3d : vec_xyz = vectors.wind_vector_3d  # m/min
    slope_vector_3d: vec_xyz = vectors.slope_vector_3d # rise/run
    # Calculate phi_W and phi_S
    # NOTE: |wind_vector_3d| = slope-aligned midflame wind speed
    phi_W: cy.float = get_phi_W(surface_fire_min, vu.vector_magnitude_3d(wind_vector_3d))
    phi_S: cy.float = get_phi_S(surface_fire_min, slope)
    # Calculate phi_E and the max_spread_direction
    phi_E_3d: vec_xyz  = get_phi_E(wind_vector_3d, slope_vector_3d, phi_W, phi_S)
    phi_E   : cy.float = vu.vector_magnitude_3d(phi_E_3d)
    max_spread_direction: vec_xyz
    if phi_E > 0.0:
        max_spread_direction = vu.as_unit_vector_3d(phi_E_3d)
    elif phi_S > 0.0:
        max_spread_direction = vu.as_unit_vector_3d(slope_vector_3d)
    else:
        max_spread_direction = (0.0, 1.0, 0.0) # default: North
    # Limit effective wind speed to max wind speed if use_wind_limit == True
    (limited_wind_speed, limited_phi_E) = maybe_limit_wind_speed(use_wind_limit,
                                                                 max_wind_speed,
                                                                 surface_fire_min,
                                                                 phi_E)
    # Calculate and return max surface fire behavior values
    max_spread_rate       : cy.float = spread_rate * (1.0 + limited_phi_E)
    max_fireline_intensity: cy.float = fireline_intensity * (1.0 + limited_phi_E)
    length_to_width_ratio : cy.float = surface_length_to_width_ratio(limited_wind_speed, surface_lw_ratio_model)
    return FireBehaviorMax(
        max_fire_type          = 1,
        max_spread_rate        = max_spread_rate,
        max_spread_direction   = max_spread_direction,
        max_fireline_intensity = max_fireline_intensity,
        max_flame_length       = calc_flame_length(max_fireline_intensity),
        length_to_width_ratio  = length_to_width_ratio,
        eccentricity           = surface_fire_eccentricity(length_to_width_ratio),
        critical_spread_rate   = 0.0,
    )
# surface-fire-behavior-max ends here
# [[file:../../org/pyretechnics.org::surface-fire-behavior-in-direction][surface-fire-behavior-in-direction]]
@cy.ccall
@cy.exceptval(check=False)
def calc_surface_fire_behavior_in_direction(surface_fire_max: FireBehaviorMax,
                                            spread_direction: vec_xyz) -> SpreadBehavior:
    """
    Given these inputs:
    - surface_fire_max     :: FireBehaviorMax struct of max surface fire behavior values
      - max_fire_type          :: 1 (surface)
      - max_spread_rate        :: m/min
      - max_spread_direction   :: (x, y, z) unit vector
      - max_fireline_intensity :: kW/m
      - max_flame_length       :: m
      - length_to_width_ratio  :: unitless (1: circular spread, > 1: elliptical spread)
      - eccentricity           :: unitless (0: circular spread, > 0: elliptical spread)
      - critical_spread_rate   :: m/min
    - spread_direction     :: 3D unit vector on the slope-tangential plane

    return a SpreadBehavior struct containing these keys:
    - dphi_dt            :: phi/min
    - fire_type          :: 1 (surface)
    - spread_rate        :: m/min
    - spread_direction   :: (x, y, z) unit vector
    - fireline_intensity :: kW/m
    - flame_length       :: m
    """
    # Unpack max surface fire behavior values
    max_spread_rate       : cy.float = surface_fire_max.max_spread_rate
    max_spread_direction  : vec_xyz  = surface_fire_max.max_spread_direction
    max_fireline_intensity: cy.float = surface_fire_max.max_fireline_intensity
    eccentricity          : cy.float = surface_fire_max.eccentricity
    # Calculate cos(w), where w is the offset angle between these unit vectors on the slope-tangential plane
    cos_w: cy.float = vu.dot_3d(max_spread_direction, spread_direction)
    # Calculate adjustment due to the offset angle from the max spread direction
    adjustment: cy.float = (1.0 - eccentricity) / (1.0 - eccentricity * cos_w)
    # Update surface fire behavior values by the adjustment value
    fireline_intensity: cy.float = max_fireline_intensity * adjustment
    return SpreadBehavior(
        dphi_dt            = 0.0,
        fire_type          = 1, # surface
        spread_rate        = max_spread_rate * adjustment,
        spread_direction   = spread_direction,
        fireline_intensity = fireline_intensity,
        flame_length       = calc_flame_length(fireline_intensity),
    )
# surface-fire-behavior-in-direction ends here



# [[file:../../org/pyretechnics.org::fuel-models-imports][fuel-models-imports]]
import cython
import cython as cy
if cython.compiled:
    from cython.cimports.libc.math import exp
    from cython.cimports.pyretechnics.cy_types import fcatarr, fclaarr, CompactFuelModel, FuelModel
else:
    from math import exp
    from pyretechnics.py_types import fcatarr, fclaarr, CompactFuelModel, FuelModel
# fuel-models-imports ends here
# [[file:../../org/pyretechnics.org::fuel-model-compact-table][fuel-model-compact-table]]
# TODO: OPTIM Replace this dictionary with something more efficient
# Lookup table including entries for each of the Anderson 13 and Scott & Burgan 40 fuel models.
#
# The fields have the following meanings:
#   {
#     fuel_model_number: (delta,
#                         M_x_dead,
#                         h,
#                         w_o_dead_1hr,
#                         w_o_dead_10hr,
#                         w_o_dead_100hr,
#                         w_o_live_herbaceous,
#                         w_o_live_woody,
#                         sigma_dead_1hr,
#                         sigma_dead_10hr,
#                         sigma_dead_100hr,
#                         sigma_live_herbaceous,
#                         sigma_live_woody), # name
#   }
compact_fuel_model_table = cy.declare(dict, { # dict[int, CompactFuelModel]
    # Anderson 13:
    # Grass and Grass-dominated (short-grass,timber-grass-and-understory,tall-grass)
    1  : (1.0, 12.0, 8.0, 0.0340, 0.0000, 0.0000, 0.0000, 0.0000, 3500.0,   0.0,  0.0,    0.0,    0.0), # R01
    2  : (1.0, 15.0, 8.0, 0.0920, 0.0460, 0.0230, 0.0230, 0.0000, 3000.0, 109.0, 30.0, 1500.0,    0.0), # R02
    3  : (2.5, 25.0, 8.0, 0.1380, 0.0000, 0.0000, 0.0000, 0.0000, 1500.0,   0.0,  0.0,    0.0,    0.0), # R03
    # Chaparral and Shrubfields (chaparral,brush,dormant-brush-hardwood-slash,southern-rough)
    4  : (6.0, 20.0, 8.0, 0.2300, 0.1840, 0.0920, 0.2300, 0.0000, 2000.0, 109.0, 30.0, 1500.0,    0.0), # R04
    5  : (2.0, 20.0, 8.0, 0.0460, 0.0230, 0.0000, 0.0920, 0.0000, 2000.0, 109.0,  0.0, 1500.0,    0.0), # R05
    6  : (2.5, 25.0, 8.0, 0.0690, 0.1150, 0.0920, 0.0000, 0.0000, 1750.0, 109.0, 30.0,    0.0,    0.0), # R06
    7  : (2.5, 40.0, 8.0, 0.0520, 0.0860, 0.0690, 0.0170, 0.0000, 1750.0, 109.0, 30.0, 1550.0,    0.0), # R07
    # Timber Litter (closed-timber-litter,hardwood-litter,timber-litter-and-understory)
    8  : (0.2, 30.0, 8.0, 0.0690, 0.0460, 0.1150, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # R08
    9  : (0.2, 25.0, 8.0, 0.1340, 0.0190, 0.0070, 0.0000, 0.0000, 2500.0, 109.0, 30.0,    0.0,    0.0), # R09
    10 : (1.0, 25.0, 8.0, 0.1380, 0.0920, 0.2300, 0.0920, 0.0000, 2000.0, 109.0, 30.0, 1500.0,    0.0), # R10
    # Logging Slash (light-logging-slash,medium-logging-slash,heavy-logging-slash)
    11 : (1.0, 15.0, 8.0, 0.0690, 0.2070, 0.2530, 0.0000, 0.0000, 1500.0, 109.0, 30.0,    0.0,    0.0), # R11
    12 : (2.3, 20.0, 8.0, 0.1840, 0.6440, 0.7590, 0.0000, 0.0000, 1500.0, 109.0, 30.0,    0.0,    0.0), # R12
    13 : (3.0, 25.0, 8.0, 0.3220, 1.0580, 1.2880, 0.0000, 0.0000, 1500.0, 109.0, 30.0,    0.0,    0.0), # R13
    # Nonburnable (NB)
    91 : (0.0,  0.0, 0.0, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000,    0.0,   0.0,  0.0,    0.0,    0.0), # NB1
    92 : (0.0,  0.0, 0.0, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000,    0.0,   0.0,  0.0,    0.0,    0.0), # NB2
    93 : (0.0,  0.0, 0.0, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000,    0.0,   0.0,  0.0,    0.0,    0.0), # NB3
    98 : (0.0,  0.0, 0.0, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000,    0.0,   0.0,  0.0,    0.0,    0.0), # NB4
    99 : (0.0,  0.0, 0.0, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000,    0.0,   0.0,  0.0,    0.0,    0.0), # NB5
    # Scott & Burgan 40:
    # Grass (GR)
    101: (0.4, 15.0, 8.0, 0.0046, 0.0000, 0.0000, 0.0138, 0.0000, 2200.0, 109.0, 30.0, 2000.0,    0.0), # GR1
    102: (1.0, 15.0, 8.0, 0.0046, 0.0000, 0.0000, 0.0459, 0.0000, 2000.0, 109.0, 30.0, 1800.0,    0.0), # GR2
    103: (2.0, 30.0, 8.0, 0.0046, 0.0184, 0.0000, 0.0689, 0.0000, 1500.0, 109.0, 30.0, 1300.0,    0.0), # GR3
    104: (2.0, 15.0, 8.0, 0.0115, 0.0000, 0.0000, 0.0872, 0.0000, 2000.0, 109.0, 30.0, 1800.0,    0.0), # GR4
    105: (1.5, 40.0, 8.0, 0.0184, 0.0000, 0.0000, 0.1148, 0.0000, 1800.0, 109.0, 30.0, 1600.0,    0.0), # GR5
    106: (1.5, 40.0, 9.0, 0.0046, 0.0000, 0.0000, 0.1561, 0.0000, 2200.0, 109.0, 30.0, 2000.0,    0.0), # GR6
    107: (3.0, 15.0, 8.0, 0.0459, 0.0000, 0.0000, 0.2479, 0.0000, 2000.0, 109.0, 30.0, 1800.0,    0.0), # GR7
    108: (4.0, 30.0, 8.0, 0.0230, 0.0459, 0.0000, 0.3352, 0.0000, 1500.0, 109.0, 30.0, 1300.0,    0.0), # GR8
    109: (5.0, 40.0, 8.0, 0.0459, 0.0459, 0.0000, 0.4132, 0.0000, 1800.0, 109.0, 30.0, 1600.0,    0.0), # GR9
    # Grass-Shrub (GS)
    121: (0.9, 15.0, 8.0, 0.0092, 0.0000, 0.0000, 0.0230, 0.0298, 2000.0, 109.0, 30.0, 1800.0, 1800.0), # GS1
    122: (1.5, 15.0, 8.0, 0.0230, 0.0230, 0.0000, 0.0275, 0.0459, 2000.0, 109.0, 30.0, 1800.0, 1800.0), # GS2
    123: (1.8, 40.0, 8.0, 0.0138, 0.0115, 0.0000, 0.0666, 0.0574, 1800.0, 109.0, 30.0, 1600.0, 1600.0), # GS3
    124: (2.1, 40.0, 8.0, 0.0872, 0.0138, 0.0046, 0.1561, 0.3260, 1800.0, 109.0, 30.0, 1600.0, 1600.0), # GS4
    # Shrub (SH)
    141: (1.0, 15.0, 8.0, 0.0115, 0.0115, 0.0000, 0.0069, 0.0597, 2000.0, 109.0, 30.0, 1800.0, 1600.0), # SH1
    142: (1.0, 15.0, 8.0, 0.0620, 0.1102, 0.0344, 0.0000, 0.1768, 2000.0, 109.0, 30.0,    0.0, 1600.0), # SH2
    143: (2.4, 40.0, 8.0, 0.0207, 0.1377, 0.0000, 0.0000, 0.2847, 1600.0, 109.0, 30.0,    0.0, 1400.0), # SH3
    144: (3.0, 30.0, 8.0, 0.0390, 0.0528, 0.0092, 0.0000, 0.1171, 2000.0, 109.0, 30.0, 1800.0, 1600.0), # SH4
    145: (6.0, 15.0, 8.0, 0.1653, 0.0964, 0.0000, 0.0000, 0.1331,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH5
    146: (2.0, 30.0, 8.0, 0.1331, 0.0666, 0.0000, 0.0000, 0.0643,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH6
    147: (6.0, 15.0, 8.0, 0.1607, 0.2433, 0.1010, 0.0000, 0.1561,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH7
    148: (3.0, 40.0, 8.0, 0.0941, 0.1561, 0.0390, 0.0000, 0.1997,  750.0, 109.0, 30.0,    0.0, 1600.0), # SH8
    149: (4.4, 40.0, 8.0, 0.2066, 0.1125, 0.0000, 0.0712, 0.3214,  750.0, 109.0, 30.0, 1800.0, 1500.0), # SH9
    # Timber-Understory (TU)
    161: (0.6, 20.0, 8.0, 0.0092, 0.0413, 0.0689, 0.0092, 0.0413, 2000.0, 109.0, 30.0, 1800.0, 1600.0), # TU1
    162: (1.0, 30.0, 8.0, 0.0436, 0.0826, 0.0574, 0.0000, 0.0092, 2000.0, 109.0, 30.0,    0.0, 1600.0), # TU2
    163: (1.3, 30.0, 8.0, 0.0505, 0.0069, 0.0115, 0.0298, 0.0505, 1800.0, 109.0, 30.0, 1600.0, 1400.0), # TU3
    164: (0.5, 12.0, 8.0, 0.2066, 0.0000, 0.0000, 0.0000, 0.0918, 2300.0, 109.0, 30.0,    0.0, 2000.0), # TU4
    165: (1.0, 25.0, 8.0, 0.1837, 0.1837, 0.1377, 0.0000, 0.1377, 1500.0, 109.0, 30.0,    0.0,  750.0), # TU5
    # Timber Litter (TL)
    181: (0.2, 30.0, 8.0, 0.0459, 0.1010, 0.1653, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL1
    182: (0.2, 25.0, 8.0, 0.0643, 0.1056, 0.1010, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL2
    183: (0.3, 20.0, 8.0, 0.0230, 0.1010, 0.1286, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL3
    184: (0.4, 25.0, 8.0, 0.0230, 0.0689, 0.1928, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL4
    185: (0.6, 25.0, 8.0, 0.0528, 0.1148, 0.2020, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0, 1600.0), # TL5
    186: (0.3, 25.0, 8.0, 0.1102, 0.0551, 0.0551, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL6
    187: (0.4, 25.0, 8.0, 0.0138, 0.0643, 0.3719, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # TL7
    188: (0.3, 35.0, 8.0, 0.2663, 0.0643, 0.0505, 0.0000, 0.0000, 1800.0, 109.0, 30.0,    0.0,    0.0), # TL8
    189: (0.6, 35.0, 8.0, 0.3053, 0.1515, 0.1905, 0.0000, 0.0000, 1800.0, 109.0, 30.0,    0.0, 1600.0), # TL9
    # Slash-Blowdown (SB)
    201: (1.0, 25.0, 8.0, 0.0689, 0.1377, 0.5051, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # SB1
    202: (1.0, 25.0, 8.0, 0.2066, 0.1951, 0.1837, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # SB2
    203: (1.2, 25.0, 8.0, 0.2525, 0.1263, 0.1377, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # SB3
    204: (2.7, 25.0, 8.0, 0.2410, 0.1607, 0.2410, 0.0000, 0.0000, 2000.0, 109.0, 30.0,    0.0,    0.0), # SB4
})
# fuel-model-compact-table ends here
# [[file:../../org/pyretechnics.org::expand-compact-fuel-model-table][expand-compact-fuel-model-table]]
# NOTE: We use this variable for X > 0.0 comparisons to account for floating point precision issues.
almost_zero = cy.declare(cy.float, 1e-6)


@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def is_burnable_fuel_model_number(fuel_model_number: cy.int) -> cy.bint:
    return not (91 <= fuel_model_number <= 99)


@cy.cfunc
@cy.exceptval(check=False)
def compute_exp_A_sigma(A: cy.float, sigma_ij: cy.float) -> cy.float:
    if sigma_ij > almost_zero:
        return exp(A / sigma_ij)
    else:
        return 0.0


@cy.cfunc
@cy.inline
@cy.exceptval(check=False)
def compute_firemod_size_class(sigma_i: cy.float) -> cy.float:
    return (
        1.0 if (sigma_i >= 1200.0)
        else 2.0 if (sigma_i >= 192.0)
        else 3.0 if (sigma_i >= 96.0)
        else 4.0 if (sigma_i >= 48.0)
        else 5.0 if (sigma_i >= 16.0)
        else 6.0
    )


@cy.cfunc
@cy.exceptval(check=False)
def expand_compact_fuel_model(fuel_model_number: cy.int) -> FuelModel:
    # Look up the CompactFuelModel by fuel_model_number
    cfm: CompactFuelModel = compact_fuel_model_table[fuel_model_number]
    # Unpack the CompactFuelModel values
    delta                : cy.float = cfm[0]
    M_x_dead             : cy.float = cfm[1]
    h                    : cy.float = cfm[2]
    w_o_dead_1hr         : cy.float = cfm[3]
    w_o_dead_10hr        : cy.float = cfm[4]
    w_o_dead_100hr       : cy.float = cfm[5]
    w_o_live_herbaceous  : cy.float = cfm[6]
    w_o_live_woody       : cy.float = cfm[7]
    sigma_dead_1hr       : cy.float = cfm[8]
    sigma_dead_10hr      : cy.float = cfm[9]
    sigma_dead_100hr     : cy.float = cfm[10]
    sigma_live_herbaceous: cy.float = cfm[11]
    sigma_live_woody     : cy.float = cfm[12]
    # Expand compressed values
    M_x_dead: cy.float = M_x_dead * 0.01
    h       : cy.float = h * 1000.0
    # Pre-compute some dynamic fuel model values
    dynamic              : cy.bint  = fuel_model_number > 100 and w_o_live_herbaceous > almost_zero
    M_x_dead_herbaceous  : cy.float = M_x_dead              if dynamic else 0.0
    sigma_dead_herbaceous: cy.float = sigma_live_herbaceous if dynamic else 0.0
    # Re-pack everything into a FuelModel struct
    return FuelModel(
        number               = fuel_model_number,
        delta                = delta,
        M_x                  = (M_x_dead,
                                M_x_dead,
                                M_x_dead,
                                M_x_dead_herbaceous,
                                0.0,
                                0.0),
        M_f                  = (0.0, 0.0, 0.0, 0.0, 0.0, 0.0),
        w_o                  = (w_o_dead_1hr,
                                w_o_dead_10hr,
                                w_o_dead_100hr,
                                0.0,
                                w_o_live_herbaceous,
                                w_o_live_woody),
        sigma                = (sigma_dead_1hr,
                                sigma_dead_10hr,
                                sigma_dead_100hr,
                                sigma_dead_herbaceous,
                                sigma_live_herbaceous,
                                sigma_live_woody),
        h                    = (h, h, h, h, h, h),
        rho_p                = (32.0, 32.0, 32.0, 32.0, 32.0, 32.0),
        S_T                  = (0.0555, 0.0555, 0.0555, 0.0555, 0.0555, 0.0555),
        S_e                  = (0.01, 0.01, 0.01, 0.01, 0.01, 0.01),
        dynamic              = dynamic,
        burnable             = is_burnable_fuel_model_number(fuel_model_number),
        exp_A_sigma          = (compute_exp_A_sigma(-138.0, sigma_dead_1hr),
                                compute_exp_A_sigma(-138.0, sigma_dead_10hr),
                                compute_exp_A_sigma(-138.0, sigma_dead_100hr),
                                compute_exp_A_sigma(-138.0, sigma_dead_herbaceous),
                                compute_exp_A_sigma(-500.0, sigma_live_herbaceous),
                                compute_exp_A_sigma(-500.0, sigma_live_woody)),
        firemod_size_classes = (compute_firemod_size_class(sigma_dead_1hr),
                                compute_firemod_size_class(sigma_dead_10hr),
                                compute_firemod_size_class(sigma_dead_100hr),
                                compute_firemod_size_class(sigma_dead_herbaceous),
                                compute_firemod_size_class(sigma_live_herbaceous),
                                compute_firemod_size_class(sigma_live_woody)),
        f_ij                 = (0.0, 0.0, 0.0, 0.0, 0.0, 0.0),
        f_i                  = (0.0, 0.0),
        g_ij                 = (0.0, 0.0, 0.0, 0.0, 0.0, 0.0),
    )


# TODO: OPTIM Replace this dictionary with something more efficient
fuel_model_table = cy.declare(dict, { # dict[int, FuelModel]
    k: expand_compact_fuel_model(k) for k in compact_fuel_model_table.keys()
})


@cy.ccall
@cy.inline
def list_fuel_model_numbers() -> list[int]:
    return list(fuel_model_table.keys())


@cy.ccall
@cy.inline
def list_fuel_models() -> list[FuelModel]:
    return list(fuel_model_table.values())


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def fuel_model_exists(fuel_model_number: cy.int) -> cy.bint:
    return fuel_model_number in fuel_model_table


@cy.ccall
@cy.inline
@cy.exceptval(check=False)
def get_fuel_model(fuel_model_number: cy.int) -> FuelModel:
    return cy.cast(FuelModel, fuel_model_table[fuel_model_number])
# expand-compact-fuel-model-table ends here
# [[file:../../org/pyretechnics.org::add-dynamic-fuel-loading][add-dynamic-fuel-loading]]
@cy.cfunc
@cy.exceptval(check=False)
def add_dynamic_fuel_loading(fuel_model: FuelModel, M_f: fclaarr) -> FuelModel:
    """
    Updates M_f and w_o. Mutates and returns fuel_model.
    """
    if fuel_model.dynamic:
        # === Dynamic Fuel Model ===
        # Calculate fraction_green and fraction_cured
        w_o                     : fclaarr  = fuel_model.w_o
        live_herbaceous_load    : cy.float = w_o[4]
        live_herbaceous_moisture: cy.float = M_f[4]
        fraction_green          : cy.float = max(0.0, min(1.0, (live_herbaceous_moisture / 0.9) - 0.3333333333333333))
        fraction_cured          : cy.float = 1.0 - fraction_green
        # Set M_f[dead_herbaceous] equal to M_f[dead_1hr]
        M_f_dynamic: fclaarr = (M_f[0], # dead_1hr
                                M_f[1],
                                M_f[2],
                                M_f[0], # dead_herbaceous
                                M_f[4],
                                M_f[5])
        # Set w_o[dead_herbaceous] and w_o[live_herbaceous] based on fraction_cured and fraction_green
        w_o_dynamic: fclaarr = (w_o[0],
                                w_o[1],
                                w_o[2],
                                live_herbaceous_load * fraction_cured, # dead_herbaceous
                                live_herbaceous_load * fraction_green, # live_herbaceous
                                w_o[5])
        # Update the passed in fuel_model to use M_f_dynamic and w_o_dynamic
        fuel_model.M_f = M_f_dynamic
        fuel_model.w_o = w_o_dynamic
        return fuel_model
    else:
        # === Static Fuel Model ===
        # Update the passed in fuel_model to use the passed in M_f
        fuel_model.M_f = M_f
        return fuel_model
# add-dynamic-fuel-loading ends here
# [[file:../../org/pyretechnics.org::add-weighting-factors][add-weighting-factors]]
# TODO: OPTIM pre-compute this conditional branching since it's fully determined by sigma.
#       This information might be represented efficiently in bit flags.
@cy.cfunc
@cy.exceptval(check=False)
def compute_gij(firemod_size_classes : fclaarr,
                f_ij                 : fclaarr,
                firemod_size_class_ij: cy.float,
                is_dead              : cy.bint) -> cy.float:
    """
    Sums the f_ij of the same category (dead/live) as i, and having the same firemod_size_class.

    NOTE: There may be repetitions in firemod_size_classes, which is why this expression is not
          trivially equal to f_ij[i].
    """
    if is_dead:
        f_ij_0: cy.float = (f_ij[0] if (firemod_size_class_ij == firemod_size_classes[0]) else 0.0)
        f_ij_1: cy.float = (f_ij[1] if (firemod_size_class_ij == firemod_size_classes[1]) else 0.0)
        f_ij_2: cy.float = (f_ij[2] if (firemod_size_class_ij == firemod_size_classes[2]) else 0.0)
        f_ij_3: cy.float = (f_ij[3] if (firemod_size_class_ij == firemod_size_classes[3]) else 0.0)
        return f_ij_0 + f_ij_1 + f_ij_2 + f_ij_3
    else:
        f_ij_4: cy.float = (f_ij[4] if (firemod_size_class_ij == firemod_size_classes[4]) else 0.0)
        f_ij_5: cy.float = (f_ij[5] if (firemod_size_class_ij == firemod_size_classes[5]) else 0.0)
        return f_ij_4 + f_ij_5


@cy.cfunc
@cy.exceptval(check=False)
def add_weighting_factors(fuel_model: FuelModel) -> FuelModel:
    """
    Assigns f_ij, f_i, and g_ij. Mutates and returns fuel_model.
    """
    # Calculate A_ij, A_i, and A_T
    w_o  : fclaarr  = fuel_model.w_o
    sigma: fclaarr  = fuel_model.sigma
    rho_p: fclaarr  = fuel_model.rho_p
    A_ij : fclaarr  = ((sigma[0] * w_o[0]) / rho_p[0],
                       (sigma[1] * w_o[1]) / rho_p[1],
                       (sigma[2] * w_o[2]) / rho_p[2],
                       (sigma[3] * w_o[3]) / rho_p[3],
                       (sigma[4] * w_o[4]) / rho_p[4],
                       (sigma[5] * w_o[5]) / rho_p[5]) # TODO: OPTIM pre-compute sigma/rho_p
    A_i_0: cy.float = A_ij[0] + A_ij[1] + A_ij[2] + A_ij[3]
    A_i_1: cy.float = A_ij[4] + A_ij[5]
    A_i  : fcatarr  = (A_i_0, A_i_1)
    A_T  : cy.float = A_i_0 + A_i_1
    # Calculate f_ij
    f_ij_0   : cy.float = 0.0
    f_ij_1   : cy.float = 0.0
    f_ij_2   : cy.float = 0.0
    f_ij_3   : cy.float = 0.0
    f_ij_4   : cy.float = 0.0
    f_ij_5   : cy.float = 0.0
    A_i_0_inv: cy.float
    A_i_1_inv: cy.float
    if A_i_0 > almost_zero:
        A_i_0_inv = 1.0 / A_i_0
        f_ij_0    = A_ij[0] * A_i_0_inv
        f_ij_1    = A_ij[1] * A_i_0_inv
        f_ij_2    = A_ij[2] * A_i_0_inv
        f_ij_3    = A_ij[3] * A_i_0_inv
    if A_i_1 > almost_zero:
        A_i_1_inv = 1.0 / A_i_1
        f_ij_4    = A_ij[4] * A_i_1_inv
        f_ij_5    = A_ij[5] * A_i_1_inv
    f_ij: fclaarr = (f_ij_0, f_ij_1, f_ij_2, f_ij_3, f_ij_4, f_ij_5)
    # Calculate f_i
    f_i_0  : cy.float = 0.0
    f_i_1  : cy.float = 0.0
    A_T_inv: cy.float
    if A_T > almost_zero:
        A_T_inv = 1.0 / A_T
        f_i_0   = A_i_0 * A_T_inv
        f_i_1   = A_i_1 * A_T_inv
    f_i: fcatarr = (f_i_0, f_i_1)
    # Calculate g_ij
    firemod_size_classes: fclaarr = fuel_model.firemod_size_classes
    g_ij                : fclaarr = (compute_gij(firemod_size_classes, f_ij, firemod_size_classes[0], True),
                                     compute_gij(firemod_size_classes, f_ij, firemod_size_classes[1], True),
                                     compute_gij(firemod_size_classes, f_ij, firemod_size_classes[2], True),
                                     compute_gij(firemod_size_classes, f_ij, firemod_size_classes[3], True),
                                     compute_gij(firemod_size_classes, f_ij, firemod_size_classes[4], False),
                                     compute_gij(firemod_size_classes, f_ij, firemod_size_classes[5], False))
    # Update the passed in fuel_model to use f_ij, f_i, and g_ij
    fuel_model.f_ij = f_ij
    fuel_model.f_i  = f_i
    fuel_model.g_ij = g_ij
    return fuel_model
# add-weighting-factors ends here
# [[file:../../org/pyretechnics.org::add-live-moisture-of-extinction][add-live-moisture-of-extinction]]
@cy.cfunc
@cy.exceptval(check=False)
def add_live_moisture_of_extinction(fuel_model: FuelModel) -> FuelModel:
    """
    Equation 88 from Rothermel 1972 adjusted by Albini 1976 Appendix III.

    Updates M_x. Mutates and returns fuel_model.
    """
    # Calculate dead_moisture_factor, dead_loading_factor, and live_loading_factor
    w_o                 : fclaarr  = fuel_model.w_o
    exp_A_sigma         : fclaarr  = fuel_model.exp_A_sigma
    M_f                 : fclaarr  = fuel_model.M_f
    M_x                 : fclaarr  = fuel_model.M_x
    M_x_dead            : cy.float = M_x[0]
    loading_factors     : fclaarr  = (w_o[0] * exp_A_sigma[0],
                                      w_o[1] * exp_A_sigma[1],
                                      w_o[2] * exp_A_sigma[2],
                                      w_o[3] * exp_A_sigma[3],
                                      w_o[4] * exp_A_sigma[4],
                                      w_o[5] * exp_A_sigma[5])
    dead_moisture_factor: cy.float = (M_f[0] * loading_factors[0] +
                                      M_f[1] * loading_factors[1] +
                                      M_f[2] * loading_factors[2] +
                                      M_f[3] * loading_factors[3])
    dead_loading_factor : cy.float = loading_factors[0] + loading_factors[1] + loading_factors[2] + loading_factors[3]
    live_loading_factor : cy.float = loading_factors[4] + loading_factors[5]
    # Calculate M_x_live
    dead_fuel_moisture: cy.float
    dead_to_live_ratio: cy.float
    M_x_live          : cy.float
    if (dead_loading_factor > almost_zero and live_loading_factor > almost_zero):
        dead_fuel_moisture = dead_moisture_factor / dead_loading_factor
        dead_to_live_ratio = dead_loading_factor / live_loading_factor
        M_x_live           = max(M_x_dead,
                                 (2.9 * dead_to_live_ratio * (1.0 - (dead_fuel_moisture / M_x_dead))) - 0.226)
    else:
        M_x_live = M_x_dead
    # Calculate M_x_new
    M_x_new: fclaarr = (M_x[0],
                        M_x[1],
                        M_x[2],
                        M_x[3],
                        M_x_live,
                        M_x_live)
    # Update the passed in fuel_model to use M_x_new
    fuel_model.M_x = M_x_new
    return fuel_model
# add-live-moisture-of-extinction ends here
# [[file:../../org/pyretechnics.org::moisturize][moisturize]]
@cy.ccall
@cy.exceptval(check=False)
def moisturize(fuel_model: FuelModel, fuel_moisture: fclaarr) -> FuelModel:
    """
    Updates w_o, M_f, and M_x and assigns f_ij, f_i, and g_ij.
    Returns a new FuelModel struct.
    """
    # Create a copy of fuel_model
    fuel_model_copy: FuelModel = FuelModel(
        number               = fuel_model.number,
        delta                = fuel_model.delta,
        M_x                  = fuel_model.M_x,
        M_f                  = fuel_model.M_f,
        w_o                  = fuel_model.w_o,
        sigma                = fuel_model.sigma,
        h                    = fuel_model.h,
        rho_p                = fuel_model.rho_p,
        S_T                  = fuel_model.S_T,
        S_e                  = fuel_model.S_e,
        dynamic              = fuel_model.dynamic,
        burnable             = fuel_model.burnable,
        exp_A_sigma          = fuel_model.exp_A_sigma,
        firemod_size_classes = fuel_model.firemod_size_classes,
        f_ij                 = fuel_model.f_ij,
        f_i                  = fuel_model.f_i,
        g_ij                 = fuel_model.g_ij,
    )
    # Mutate and return the copy
    dynamic_fuel_model    : FuelModel = add_dynamic_fuel_loading(fuel_model_copy, fuel_moisture)
    weighted_fuel_model   : FuelModel = add_weighting_factors(dynamic_fuel_model)
    moisturized_fuel_model: FuelModel = add_live_moisture_of_extinction(weighted_fuel_model)
    return moisturized_fuel_model
# moisturize ends here
